{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "59b49a34",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "# Data Classes and Validation\n",
    "\n",
    "## Data Classes\n",
    "\n",
    "### dataclasses\n",
    "\n",
    "Python 3.7 introduced the [@dataclass](https://docs.python.org/3/library/dataclasses.html) decorator to write\n",
    "classes meant to hold data with ease and minimal repetition. Consider the following\n",
    "example from our last chapter."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "12294952",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "class Person:\n",
    "    def __init__(self, birthday, name):\n",
    "        self.birth_day = birthday[0]\n",
    "        self.birth_month = birthday[1]\n",
    "        self.birth_year = birthday[2]\n",
    "        self.name = name"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "ec2af4fc",
   "metadata": {},
   "source": [
    "A good practice here would be to add other dunder methods like `__repr__` and `__eq__`\n",
    "for a better user experience. Let's do that."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e90e768f",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "class Person:\n",
    "    def __init__(self, birth_day, birth_month, birth_year, name):\n",
    "        self.birth_day = birth_day[0]\n",
    "        self.birth_month = birth_month[1]\n",
    "        self.birth_year = birth_year[2]\n",
    "        self.name = name\n",
    "    def __repr__(self):\n",
    "        return f\"Person(birth_day={self.birth_day}, birth_month={self.birth_month}, birth_year={self.birth_year}, name='{self.name}')\"\n",
    "    def __eq__(self, other):\n",
    "        return (\n",
    "            self.birth_day == other.birth_day\n",
    "            and self.birth_month == other.birth_month\n",
    "            and self.birth_year == other.birth_year\n",
    "            and self.name == other.name\n",
    "        )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "52a73cf5",
   "metadata": {},
   "source": [
    "Simple enough! But can we make it even simpler and DRYer? Yes!\n",
    "The `@dataclass` decorator automates the creation of a the `__init__`,\n",
    "`__repr__`, and `__eq__` methods."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ee5d3b5a",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "from dataclasses import dataclass\n",
    "\n",
    "@dataclass\n",
    "class Person:\n",
    "    birth_day: int\n",
    "    birth_month: int\n",
    "    birth_year: int\n",
    "    name: str"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b3966f3b",
   "metadata": {},
   "source": [
    "That's it! Data classes rely heavily on static types. Let us try creating the objects and using the methods."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3aaf7ab7",
   "metadata": {},
   "outputs": [],
   "source": [
    "p1 = Person(7, 1, 2002, \"Saransh\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "71b8233c",
   "metadata": {},
   "outputs": [],
   "source": [
    "p2 = Person(16, 3, 2002, \"Hnin\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "7716b5d2",
   "metadata": {},
   "outputs": [],
   "source": [
    "p1, p2"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "11cdbfa0",
   "metadata": {},
   "outputs": [],
   "source": [
    "p1 == p2"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "760394ac",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "p2 == p2"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "4fe39964",
   "metadata": {},
   "source": [
    "As regular classes, one can assign default values to the parameters."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "b7afbeab",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "@dataclass\n",
    "class Person:\n",
    "    birth_day: int = 1\n",
    "    birth_month: int = 1\n",
    "    birth_year: int = 1972\n",
    "    name: str = \"John Doe\""
   ]
  },
  {
   "cell_type": "markdown",
   "id": "7576145c",
   "metadata": {},
   "source": [
    "To have more flexibility, one can use the `field` function and pass in arguments\n",
    "such as `default` (default value), `repr` (whether to include in repr),\n",
    "`init` (whether to include in init), `compare` (whether to include in eq), etc."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d4a715ae",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "from dataclasses import field\n",
    "\n",
    "@dataclass\n",
    "class Person:\n",
    "    birth_day: int = field(default=1)\n",
    "    birth_month: int = field(default=1)\n",
    "    birth_year: int = field(default=1972)\n",
    "    name: str = field(default=\"John Doe\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "bfbfffc8",
   "metadata": {},
   "source": [
    "We can extend out example to include a `People` class that takes in a `list`\n",
    "of `Person`s and has the same basic dunder methods."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "341a2b91",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "from typing import List\n",
    "\n",
    "@dataclass\n",
    "class People:\n",
    "   person: List[Person] = []"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "988136ff",
   "metadata": {},
   "source": [
    "Why does that throw an error? Using mutable objects as function argument defaults\n",
    "is a bad practice as they will be edited in-place inside the function definition.\n",
    "The argument variable is shared throughout all the calls to a function, which\n",
    "is known as [\"call by sharing.\"](https://en.wikipedia.org/wiki/Evaluation_strategy)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3b570a39",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "def append_to_list(el, lst=[]):\n",
    "    lst.append(el)\n",
    "    return lst"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "51b7faa5",
   "metadata": {},
   "outputs": [],
   "source": [
    "append_to_list(0)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d59821d2",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "append_to_list(1)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "063d19ac",
   "metadata": {},
   "source": [
    "The error can be fixed by using the `default_factory` parameter\n",
    "of the `fields` function."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3506f75b",
   "metadata": {},
   "outputs": [],
   "source": [
    "@dataclass\n",
    "class People:\n",
    "   person: List[Person] = field(default_factory=list)\n",
    "\n",
    "People([p1, p2])"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e13ece06",
   "metadata": {},
   "source": [
    "Finally, dataclasses offer several other features to the users, such as -\n",
    "- ability to override the generated dunder methods\n",
    "- a `__post_init__` method for data validation\n",
    "- numerous arguments for the decorator -\n",
    "  - order: generate methods for comparison operators\n",
    "  - frozen: add no setter for the fields\n",
    "  - ...\n",
    "\n",
    "Moreover, data classes are syntactic sugars to make your life easier when\n",
    "building classes that store data. They might not be very useful if you\n",
    "want the dunder methods to behave in a specific way, as you will end up\n",
    "overriding the generated methods. Additionally, given the `dataclasses`\n",
    "is part of the Python standard library, addition of new features and bug\n",
    "fixes are relatively slow, but it blends really well with Python and the\n",
    "fundamental PyData libraries. [attrs](https://www.attrs.org/en/stable/)\n",
    "is a third-party solution that allows one to write concise classes without\n",
    "much repetition.\n",
    "\n",
    "## Data Validation\n",
    "\n",
    "Any kind of data requires proper validation before it can be used to execute\n",
    "actual science. Data validation allows your data pipelines to \"fail fast\" or\n",
    "fail at the very first step, saving compute resources and time. One can either\n",
    "define custom rules in `__init__` or `__post_init__`, or use libraries\n",
    "like pydantic to automate the process.\n",
    "\n",
    "### Pydantic\n",
    "\n",
    "[Pydantic](https://docs.pydantic.dev/latest/) enables data validation using Python\n",
    "type hints and class inheritance."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "32a79811",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "from pydantic import BaseModel, Field\n",
    "\n",
    "class Person(BaseModel):\n",
    "    birth_day: int = Field(1)\n",
    "    birth_month: int = Field(1)\n",
    "    birth_year: int = Field(1972)\n",
    "    name: str = Field(\"John Doe\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b1d0110c",
   "metadata": {},
   "source": [
    "BaseModel subclasses take keyword arguments only. The `Field` class offers\n",
    "several powerful arguments, like -\n",
    "- examples: example values for the field\n",
    "- description: desription of the field\n",
    "- frozen: removes setters for the fields\n",
    "- init, repr, and compare as dataclass\n",
    "- ...."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0c39753a",
   "metadata": {},
   "outputs": [],
   "source": [
    "data = {\"birth_day\": 7, \"birth_month\": 1, \"birth_year\": 2002, \"name\": \"Saransh\"}\n",
    "Person(**data)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b696a35e",
   "metadata": {},
   "source": [
    "Works like a data class!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d6e626eb",
   "metadata": {},
   "outputs": [],
   "source": [
    "p1 = Person(**{\"birth_day\": 7, \"birth_month\": 1, \"birth_year\": 2002, \"name\": \"Saransh\"})\n",
    "p2 = Person(**{\"birth_day\": 16, \"birth_month\": 3, \"birth_year\": 2002, \"name\": \"Hnin\"})\n",
    "\n",
    "p1, p2"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "75503095",
   "metadata": {},
   "outputs": [],
   "source": [
    "p1 == p2"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d1fb5105",
   "metadata": {},
   "outputs": [],
   "source": [
    "p1 == p1"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "7321b1d2",
   "metadata": {},
   "source": [
    "But it can validate data using type hints."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "08c1d9f0",
   "metadata": {},
   "outputs": [],
   "source": [
    "Person.model_validate(data)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ce4b33c2",
   "metadata": {},
   "outputs": [],
   "source": [
    "bad_data = {\"birth_day\": 7, \"birth_month\": 1, \"birth_year\": 2002, \"name\": 1}\n",
    "\n",
    "Person.model_validate(bad_data)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "07e583e6",
   "metadata": {},
   "source": [
    "The library further offers custom types, such as `EmailStr` for email address\n",
    "validation or `SecretStr` which hides sensitive information. Users can also\n",
    "write their own validation scripts using the `@field_validator` decorator."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "2c6b8374",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "from pydantic import field_validator\n",
    "\n",
    "class Person(BaseModel):\n",
    "    birth_day: int = Field(1)\n",
    "    birth_month: int = Field(1)\n",
    "    birth_year: int = Field(1972)\n",
    "    name: str = Field(\"John Doe\")\n",
    "    \n",
    "    @field_validator(\"name\")\n",
    "    @classmethod\n",
    "    def validate_name(cls, v: str) -> str:\n",
    "        if \" \" not in v:\n",
    "            raise ValueError(\"the name must have first and last names separated by <space>\")\n",
    "        return v"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "242fe9c8",
   "metadata": {},
   "source": [
    "Similarly, the whole data model or connect multiple fields together for\n",
    "validation using `@model_validator`. The \"after\" mode tell pydantic to run\n",
    "its own validations on the data before running the custom validations\n",
    "defined by the user."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "dfc195a1",
   "metadata": {},
   "outputs": [],
   "source": [
    "from pydantic import model_validator\n",
    "from typing import Any\n",
    "\n",
    "class Person(BaseModel):\n",
    "    birth_day: int = Field(1)\n",
    "    birth_month: int = Field(1)\n",
    "    birth_year: int = Field(1972)\n",
    "    name: str = Field(\"John Doe\")\n",
    "    \n",
    "    @model_validator(mode=\"after\")\n",
    "    @classmethod\n",
    "    def validate_name(self, v: Any) -> Person:\n",
    "        if \" \" not in v[\"name\"]:\n",
    "            raise ValueError(\"the name must have first and last names separated by <space>\")\n",
    "        if v[\"birth_month\"] < 1 or v[\"birth_month\"] > 12:\n",
    "            raise ValueError(\"birth month should be between 1 and 12\")\n",
    "        if v[\"birth_month\"] in [1, 3, 5, 7, 8, 10, 12] and (v[\"birth_day\"] < 1 or v[\"birth_day\"] > 31):\n",
    "            raise ValueError(\"birth day should be between 1 and 31\")\n",
    "        if v[\"birth_month\"] in [4, 6, 9, 11] and (v[\"birth_day\"] < 1 or v[\"birth_day\"] > 31):\n",
    "            raise ValueError(\"birth day should be between 1 and 30\")\n",
    "        if v[\"birth_month\"] == 2:\n",
    "            if (v[\"birth_year\"] % 400 == 0 or v[\"birth_year\"] % 4 == 0):\n",
    "                if v[\"birth_day\"] < 1 or v[\"birth_day\"] > 29:\n",
    "                    raise ValueError(\"birth day should be between 1 and 29\")\n",
    "            else:\n",
    "                    raise ValueError(\"birth day should be between 1 and 28\")\n",
    "        return v"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "784f370c",
   "metadata": {},
   "source": [
    "Pydantic has a ton of other features and functionalities that make data processing\n",
    "for every domain easy and simple. Overall, static typing facilitates data classes\n",
    "and pydantic validations. Several libraries build on top of pydantic to provide\n",
    "domain specific validation tools. For example [BPX](https://github.com/FaradayInstitution/BPX)\n",
    "implements the Battery Parameter eXchange (BPX) schema in Pydantic with validation\n",
    "parsing, and JSON serialization capabilities."
   ]
  }
 ],
 "metadata": {
  "jekyll": {
   "display_name": "Data Classes and Validation"
  },
  "jupytext": {
   "main_language": "python",
   "notebook_metadata_filter": "-kernelspec,jupytext,jekyll"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}